#include "Block.h"
#include "SDL2_gfxPrimitives.h"

/**
 * @brief Block ctor
 * @param x x position, horizontal index within tetris window
 * @param y y position, vertical index within tetris window
 * @param colour - colour in hexadecimal format 0xRRGGBBAA
 * @param containing_window - the tetris window the block is in
 */
Block::Block(int x, int y, Uint32 colour, const TetrisWindow* containing_window)
    : x(x)
    , y(y)
    , colour(colour)
    , window(containing_window) {
}

/**
 * @brief Displays the block on the screen using the global renderer
 * @details Gets actual pixel coordinates and sizes from containing window
 *          based on the relative x and y position stored
 */
void Block::draw() const {
    int size = window->get_block_size();
    int x1 = window->get_x1(x);
    int y1 = window->get_y1(y);
    boxColor(sdlRenderer, x1+1, y1+1, x1+size-1, y1+size-1, colour);
}

/**
 * @brief Check if block is movable in a certain direction
 * @param direction - direction the block is trying to be moved in (-1 left, 0 down, 1 right)
 * @retval true the position in the given direction is available
 * @retval false movement of the block is blocked
 */
bool Block::movable(short direction) const {    
    switch(direction) {
        case -1:
            return window->is_available(x-1, y);
            break;
        case 0:
            return window->is_available(x, y+1);
            break;
        case 1:
            return window->is_available(x+1, y);
            break;
        default:
            std::cerr << "Direction error in movable()" << std::endl;
            return false;
    }
}

/**
 * @brief Moves the block
 * @param direction direction in which to move the block (-1 left, 0 down, 1 right)
 * @retval true move successful
 * @retval false move unsuccessful
 */
bool Block::move(short direction) {
    /* Safety check to make sure the block is movable */
    if(!movable(direction)) {
        return false;
    }
    
    switch(direction) {
        case -1:
            x--;
            break;
        case 0:
            y++;
            break;
        case 1:
            x++;
            break;
        default:
            std::cerr << "Direction error in move()" << std::endl;
    }
    return true;
}

/**
 * @brief Check if block is in a certain position
 * @param x horizontal position in tetris window
 * @param y vertical position in tetris window
 * @retval true block is in specified position
 * @retval false block is not in specified position
 */
bool Block::is_at(int x, int y) const {
    return x == this->x && y == this->y;
}

/**
 * @brief x position getter
 * @return the x position of the block
 */
int Block::get_x() const {
    return x;
}

/**
 * @brief y position getter
 * @return the y position of the block
 */
int Block::get_y() const {
    return y;
}

/**
 * @brief Position setter
 * @param x new x position to set
 * @param y new y position to set
 */
void Block::set(int x, int y) {
    this->x = x;
    this->y = y;
}
