#include "SoundPlayer.h"

/**
 * @brief SoundPlayer ctor
 * @param filename name of the sound file to play
 * @details Intializes SDL_AUDIO and Mix_OpenAudio - be aware that I did
 *          not find a way to check for these being already initialized
 *          separately, don't use this if you have other parts of your
 *          program using the SDL_mixer (but it is compatible with MusicPlayer)
 */
SoundPlayer::SoundPlayer(std::string filename) : filename(filename) {
    /* Check if SDL_AUDIO is already running */
    if(SDL_WasInit(SDL_INIT_AUDIO) == 0) {
        /* Initialize SDL_AUDIO */
        if(SDL_Init(SDL_INIT_AUDIO) < 0) {
            std::cerr << "Failed to init SDL audio: " << SDL_GetError() << std::endl;
        }
        /* Initialize Mixer */
        if(Mix_OpenAudio(44100, MIX_DEFAULT_FORMAT, 2, 2048) < 0) {
            std::cerr << "Failed to init SDL mixer: " << Mix_GetError() << std::endl;
        }
    }
    
    /* Load sound file */
    sound = Mix_LoadWAV(filename.c_str());
    if(sound == nullptr) {
        std::cerr << "Failed to load sound \"" << filename << "\": " << Mix_GetError() << std::endl;
    }
}

/**
 * @brief SoundPlayer dtor, frees contained sound
 */
SoundPlayer::~SoundPlayer() {
    Mix_FreeChunk(sound);
}

/**
 * @brief Filename getter
 * @return name of the sound file stored
 */
std::string SoundPlayer::get_filename() const {
    return filename;
}

/**
 * @brief Plays the sound once
 */
void SoundPlayer::play() {
    Mix_PlayChannel(-1, sound, 0);
}

/**
 * @brief Decreases volume by 16 units (8 total steps, range is 0-128)
 */
void SoundPlayer::volume_down() {
    int current_volume = Mix_Volume(-1, -1);
    if(current_volume > 0) {
        Mix_Volume(-1, current_volume-16<0 ? 0 : current_volume-16);
    }
}

/**
 * @brief Increases volume by 16 units (8 total steps, range is 0-128)
 */
void SoundPlayer::volume_up() {
    int current_volume = Mix_Volume(-1, -1);
    Mix_Volume(-1, current_volume+16);
}
