#include "TetrisWindow.h"
#include "IndividualPieces.h"

/**
 * @brief Standard SDL timer function
 * @param ms time between generated events, milliseconds
 * @param param empty parameter, required by SDL
 * @return ms, required by SDL
 */
Uint32 timer(Uint32 ms, void* param) {
    SDL_Event event;
    event.type = SDL_USEREVENT;
    SDL_PushEvent(&event);
    return ms;
}

/**
 * @brief Creates a new random piece for the game (allocates memory!)
 * @return The new piece
 */
Piece* TetrisWindow::random_piece() const {
    /* hori_blocks/2 initializes new piece to be
     * in the middle of the tetris window */
    switch(rand()%7) {
        case 0:
            return new Piece_I(hori_blocks/2, this);
        case 1:
            return new Piece_J(hori_blocks/2, this);
        case 2:
            return new Piece_L(hori_blocks/2, this);
        case 3:
            return new Piece_O(hori_blocks/2, this);
        case 4:
            return new Piece_S(hori_blocks/2, this);
        case 5:
            return new Piece_T(hori_blocks/2, this);
        default:
            return new Piece_Z(hori_blocks/2, this);
    }
}

/**
 * @brief TetrisWindow ctor
 * @param x_pos pixels from left side of screen
 * @param y_pos pixels from top of screen
 * @param horizontal_blocks number of blocks in the horizontal direction
 * @param vertical_blocks number of blocks in the vertical direction
 * @param block_size the size of a single block
 * @param music whether music should be played ingame
 */
TetrisWindow::TetrisWindow(unsigned x_pos, unsigned y_pos, unsigned horizontal_blocks, unsigned vertical_blocks, unsigned block_size, bool music)
    : x(x_pos)
    , y(y_pos)
    , vert_blocks(vertical_blocks)
    , hori_blocks(horizontal_blocks)
    , block_size(block_size)
    , score(0)
    , sound_enabled(music) {
    /* Creates empty vectors for each row */
    for(unsigned i = 0; i < vertical_blocks; i++) {
        std::vector<Block*> temp;
        rows.push_back(temp);
    }
}

/**
 * @brief Frees memory that it's responsible for
 */
TetrisWindow::~TetrisWindow() {
    for(auto b : blocks) {
        delete b;
    }
}

/**
 * @brief Display TetrisWindow using the global renderer
 */
void TetrisWindow::draw() const {
    /* Background */
    boxColor(sdlRenderer, x, y, x+hori_blocks*block_size, y+vert_blocks*block_size, 0x444444FF);
    /* Vertical lines */
    for(unsigned i = 0; i <= hori_blocks; i++) {
        lineColor(sdlRenderer, x+i*block_size, y, x+i*block_size, y+vert_blocks*block_size, 0x666666FF);
    }
    /* Horizontal lines */
    for(unsigned i = 0; i <= vert_blocks; i++) {
        lineColor(sdlRenderer, x, y+i*block_size, x+hori_blocks*block_size, y+i*block_size, 0x666666FF);
    }
    /* Blocks owned by the window */
    for(auto b : blocks) {
        b->draw();
    }
    /* Blocks owned by the current moving piece */
    if(current_piece != nullptr) {
        current_piece->draw();
    }
}

/**
 * @brief Checks if a certain spot is available
 * @param x x position within window
 * @param y y position within window
 * @retval true specified position is available
 * @retval false specified position is occupied or out of bounds
 */
bool TetrisWindow::is_available(unsigned x, unsigned y) const {
    if(x < 0 || y < 0 || x > hori_blocks-1 || y > vert_blocks-1) {
        return false;
    }
    
    for(auto b : blocks) {
        if(b->is_at(x, y)) {
            return false;
        }
    }
    
    return true;
}

/**
 * @brief Event loop for the main game
 * @return the score achieved by the player
 */
unsigned TetrisWindow::loop() {
    bool quit = false; /* Flag for breaking the loop */
    bool game_over = false; /* Flag if quit happened because the player failed */
    bool paused = false; /* Flag for game being paused */
    
    int timer_delay = 333; /* Delay between two drops of the piece, in ms */
    
    srand(time(0)); /* Seed the random number generator */
    
    SDL_TimerID sdlTimer;
    
    MusicPlayer mp("OctoNote - Korobeiniki Tetris Remix (V2).ogg"); /* Load music */
    SoundPlayer sp_row_cleared("126426__cabeeno-rossley__timer-ends-time-up.wav"); /* Load row cleared sound effect */
    /* Start music if sound is enabled */
    if(sound_enabled) {
        mp.play();
    }
    
    /* Start timer */
    sdlTimer = SDL_AddTimer(timer_delay, timer, NULL);
    /* Get first piece */
    current_piece = random_piece();
    
    while(SDL_WaitEvent(&ev) && !quit && !game_over) {
        /* Normal game loop */
        if(!paused) {
            switch(ev.type) {
                case SDL_KEYDOWN:
                    switch(ev.key.keysym.sym) {
                        case SDLK_LEFT: /* Move left */
                            current_piece->move(-1);
                            break;
                        case SDLK_RIGHT: /* Move right */
                            current_piece->move(1);
                            break;
                        case SDLK_UP: /* Rotate */
                            current_piece->rotate();
                            break;
                        case SDLK_DOWN: /* Drop */
                            /* Move current piece down as many times as possible */
                            while(current_piece->move(0));
                            /* If the piece can't move anymore, and hasn't moved at all */
                            if(current_piece->get_move_count() == 0) {
                                game_over = true;
                            }
                            else if(replace_current_piece()) {
                                /* If row has been cleared while placing down current piece,
                                 * play sound effect */
                                if(sound_enabled) {
                                    sp_row_cleared.play();
                                }
                            }
                            break;
                            
                        case SDLK_i: /* Music volume up */
                            mp.volume_up();
                            break;
                        case SDLK_k: /* Music volume down */
                            mp.volume_down();
                            break;
                            
                        case SDLK_u: /* Sound effect volume up */
                            sp_row_cleared.volume_up();
                            break;
                        case SDLK_j: /* Sound effect volume down */
                            sp_row_cleared.volume_down();
                            break;
                            
                        case SDLK_SPACE: /* Pause */
                            /* Pause the game by removing the timer */
                            SDL_RemoveTimer(sdlTimer);
                            paused = true;
                            break;
                    }
                    break;
                    
                case SDL_USEREVENT: /* Timer event, progress game */
                    /* Try moving current piece down */
                    if(! current_piece->move(0)) {
                        /* If piece can't be moved, and it hasn't moved a single time -
                         * the screen is filled, game over */
                        if(current_piece->get_move_count() == 0) {
                            game_over = true;
                        }
                        /* If piece can't be moved any more, but has moved at least one time,
                         * it has landed on top of other pieces - replace it */
                        else if(replace_current_piece()) {
                            /* If row has been cleared while placing down current piece,
                             * play sound effect */
                            if(sound_enabled) {
                                sp_row_cleared.play();
                            }
                        }
                    }
                    break;
                    
                case SDL_QUIT: /* Exit */
                    quit = true;
                    break;
            }
            
            /* Redraw window and contents */
            draw();
            SDL_RenderPresent(sdlRenderer);
        }
        /* In case the game is paused */
        else {
            switch(ev.type) {
                case SDL_KEYDOWN:
                    switch(ev.key.keysym.sym) {
                        case SDLK_SPACE: /* Resume */
                            /* Resume game by starting the timer again */
                            sdlTimer = SDL_AddTimer(timer_delay, timer, NULL);
                            paused = false;
                            break;
                    }
                    break;
                case SDL_QUIT: /* Exit */
                    quit = true;
                    break;
            }
        }
        
        
    }
    
    /* Free current piece */
    delete current_piece;
    /* Stop timer */
    SDL_RemoveTimer(sdlTimer);
    
    /* If the player failed */
    if(game_over) {
        /* Save current volume, set it very low */
        unsigned mp_volume = mp.volume_get();
        mp.volume_set(6);
        /* Get size of TetrisWindow */
        unsigned h = vert_blocks*block_size;
        unsigned w = hori_blocks*block_size;
        /* Fade window */
        boxColor(sdlRenderer, x, y, x+w, y+h, 0x293134DD);
        /* Display texts */
        centeredtextColor(sdlRenderer, x+w/2, y+h/2, 60, "GAME OVER", 0x3494BAFF);
        centeredtextColor(sdlRenderer, x+w/2, y+h/2+40, 16, "Press any key to return to the menu", 0x3494BAFF);
        /* Submit changes to screen */
        SDL_RenderPresent(sdlRenderer);
        /* Wait until mouse click, key down, or quit event occurs */
        while(SDL_WaitEvent(&ev) && ev.type != SDL_KEYDOWN &&
                ev.type != SDL_MOUSEBUTTONDOWN && ev.type != SDL_QUIT);
        /* Restore volume to previous level */
        mp.volume_set(mp_volume);
    }
    
    /* Stop the ingame music */
    mp.stop();
    
    return score;
}

/**
 * @brief Block size getter function
 * @return the size of a single block
 */
unsigned TetrisWindow::get_block_size() const {
    return block_size;
}

/**
 * @brief x position getter for Block
 * @param x x coordinate of Block
 * @return The x position where the Block should be drawn, in pixels
 */
unsigned TetrisWindow::get_x1(unsigned x) const {
    return this->x + block_size*x;
}

/**
 * @brief y position getter for Block
 * @param y y coordinate of Block
 * @return The y position where the Block should be drawn, in pixels
 */
unsigned TetrisWindow::get_y1(unsigned y) const {
    return this->y + block_size*y;
}

/**
 * @brief Dumps current moving piece, gets a new one
 * @details Also checks for any completed rows and manages all that
 * @retval true at least one row has been completed and cleared
 * @retval false no rows have been completed
 */
bool TetrisWindow::replace_current_piece() {
    /* Row complete flag */
    bool row_cleared = false;
    /* Dumps the blocks owned by the currently moving piece
     * into the collection of the TetrisWindow's Blocks */
    while(current_piece->has_blocks()) {
        blocks.push_back(current_piece->get_block());
        /* After adding to blocks, also adds to appropriate row */
        rows[blocks.back()->get_y()].push_back(blocks.back()); /* Huh? */
    }
    /* Frees memory of the current piece */
    delete current_piece;
    /* Gets a new randomly selected piece, memory allocation happens here */
    current_piece = random_piece();
    
    /* Iterate through rows from the bottom */
    int i = vert_blocks;
    while(i --> 0) {
        /* If row is complete */
        if(completed(rows[i])) {
            /* Increase score, set flag */
            score += 100;
            row_cleared = true;
            
            /* Remove the Blocks in the row from the blocks vector */
            for(Block* b : rows[i]) {
                blocks.erase(std::find(blocks.begin(), blocks.end(), b));
                delete b;
            }
            /* Remove all Blocks from the row */
            rows[i].clear();
            /* Move the Blocks in all the above rows down one unit */
            int j = i;
            while(j --> 0) {
                for(Block* b : rows[j]) {
                    b->move(0);
                }
            }
            /* Row indexes are now off, shift each row one over */
            int k = vert_blocks;
            while(k --> 1) {
                rows[k] = rows[k-1];
            }
            /* Don't progress the loop - the current row can be
             * completed again, it needs checking */
            i++;
        }
    }
    
    return row_cleared;
}

/**
 * @brief Check if a row is completed
 * @param row the Blocks in the row
 * @retval true row is completed
 * @retval false row is incomplete
 */
bool TetrisWindow::completed(std::vector<Block*> row) {
    /* This is not the most sophisticated way of doing this,
     * but it's quick and it works well */
    return row.size() == hori_blocks;
}
